import numpy as np
from numpy.random import Generator
import xarray as xr
import uxarray as uxr
from scipy.optimize import fsolve
from typing import Tuple, Any
from numpy.typing import ArrayLike
from .target import Target
from ..utils.custom_types import FloatLike
from ..utils import montecarlo as mc
from .surface import Surface
from .target import Target
from ..cython import morphology

RIMDROP = 4.20

class Morphology:
    """
    An operations class for computing the morphology of a crater based on its size and target properties.

    This class encapsulates the logic for altering the topography of the surface based on the crater properties.

    Parameters
    ----------
    crater : Crater
        The crater to be created.
    target : Target
        The target body for the impact simulation.
    surf : Surface
        The surface to be altered.
    ejecta_truncation : float, optional
        The relative distance from the rim of the crater to truncate the ejecta blanket, default is None, which will compute a 
        truncation distance based on where the ejecta thickness reaches a small value.         
    rng : Generator, optional
        A random number generator instance. If not provided, the default numpy RNG will be used.
    """
    
    def __init__(self, 
                 crater,  
                 target: Target | None = None, 
                 rng: Generator | None = None,
                 ejecta_truncation: FloatLike | None = None
                 ):

        self.crater = crater 
        self.target = target
        self.rng = rng
        self.ejecta_truncation = ejecta_truncation

        # Set the morphology based on crater type
        self.set_morphology_parameters()
       
    def __repr__(self):
        return (f"Morphology(morphology_type={self.morphology_type}, diameter={self.diameter}, "
                f"rimheight: {self.rimheight}, rimwidth: {self.rimwidth}, floordepth: {self.floordepth}, floordiam: {self.floordiam})") 
    
    def set_morphology_parameters(self):
        """
        Sets the morphology parameters based on the crater type.
        """
        diameter_km = self.diameter * 1e-3  # Convert to km for these models

        if self.morphology_type in ["simple", "transitional"]:
            # A hybrid model between Pike (1977) and Fassett & Thomson (2014)
            self.rimheight = 0.043 * diameter_km**1.014 * 1e3  # Closer to Fassett & Thomson
            self.rimwidth = 0.257 * diameter_km**1.011 * 1e3   # Pike model
            self.floordepth = 0.224 * diameter_km**1.010 * 1e3 # Closer to Fassett & Thomson
            self.floordiam = 0.200 * diameter_km**1.143 * 1e3  # Fassett & Thomson for D~1km, Pike for D~20km

        elif self.morphology_type in ["complex", "peakring", "multiring"]:
            # Following Pike (1977)
            self.rimheight = 0.236 * diameter_km**0.399 * 1e3  # Pike model
            self.rimwidth = 0.467 * diameter_km**0.836 * 1e3   # Pike model
            self.floordepth = 1.044 * diameter_km**0.301 * 1e3 # Pike model
            # Fassett & Thomson for D~1km, Pike for D~20km, but limited to 90% of diameter
            self.floordiam = min(0.187 * diameter_km**1.249 * 1e3, 0.9 * self.diameter)
            self.peakheight = 0.032 * diameter_km**0.900 * 1e3  # Pike model
        else:
            raise ValueError(f"Unknown morphology type: {self.morphology_type}")
            
        self.ejrim = 0.14 * (self.diameter * 0.5)**(0.74) # McGetchin et al. (1973) Thickness of ejecta at rim

    def profile(self, r: ArrayLike, r_ref: ArrayLike) -> np.float64:
        elevation = morphology.profile(r,
                                       r_ref, 
                                       self.diameter, 
                                       self.floordepth, 
                                       self.floordiam, 
                                       self.rimheight, 
                                       self.ejrim, 
                                       RIMDROP)
         
        return np.array(elevation, dtype=np.float64)
    
    def form_crater(self, 
                    surf: Surface,
                    **kwargs) -> None:
        """
        This method forms the interior of the crater by altering the elevation variable of the surface mesh.
        
        Parameters
        ----------
        surf : Surface
            The surface to be altered.
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions (not used here).
        """
      
        rmax = self.compute_rmax(minimum_thickness=surf.smallest_length)
        region_surf = surf.extract_region(self.crater.location, rmax)
        if not region_surf: # The crater is too small to change the surface
            return

        region_surf['node_crater_distance'], region_surf['face_crater_distance'] = region_surf.get_distance(self.crater.location)
        #region_surf['node_crater_bearing'], region_surf['face_crater_bearing']  = surf.get_initial_bearing(self.crater.location)
        region_surf.get_reference_surface(self.crater.location, self.crater.radius)
        
        try:
            node_elevation = self.profile(region_surf['node_crater_distance'].values, region_surf['reference_node_elevation'].values)
            surf['node_elevation'].loc[{'n_node': region_surf.uxgrid._ds["subgrid_node_indices"]}] = node_elevation
            
            face_elevation = self.profile(region_surf['face_crater_distance'].values, region_surf['reference_face_elevation'].values)
            surf['face_elevation'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] = face_elevation
        except:
            print(self)
            raise ValueError("Something went wrong with this crater!")
                 
        return  
   
    def compute_rmax(self,
                     minimum_thickness: np.float64,
                     ) -> np.float64:
        """
        Compute the maximum extent of the crater based on the minimum thickness of the ejecta blanket or the ejecta_truncation factor,
        whichever is smaller.
        
        Parameters
        ----------
        minimum_thickness : np.float64
            The minimum thickness of the ejecta blanket in meters.
        """ 
        
        # Compute the reference surface for the crater 
        def _profile_invert(r):
            return self.profile(r, np.zeros(1)) - minimum_thickness
       
        # Get the maximum extent 
        rmax = fsolve(_profile_invert, x0=self.radius)[0]
        if self.ejecta_truncation:
            rmax = min(rmax, self.ejecta_truncation * self.radius)        
        # Get the maximum extent 
        rmax = fsolve(_profile_invert, x0=self.radius)[0]
        if self.ejecta_truncation:
            rmax = min(rmax, self.ejecta_truncation * self.radius)
            
        return rmax

    @property
    def diameter(self) -> np.float64:
        """
        The diameter of the crater in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.diameter

    @property
    def radius(self) -> np.float64:
        """
        The radius of the crater in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.radius
   
    @property 
    def morphology_type(self) -> str:
        """
        The morphology type of the crater.
        
        Returns
        -------
        str 
        """
        return self.crater.morphology_type 
    
    @property
    def rimheight(self) -> np.float64:
        """
        The height of the crater rim in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.rimheight
    
    @rimheight.setter
    def rimheight(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("rimheight must be of type FloatLike")
        self.crater.rimheight = np.float64(value)
        
    @property
    def rimwidth(self) -> np.float64:
        """
        The width of the crater rim in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.rimwidth
    
    @rimwidth.setter
    def rimwidth(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("rimwidth must be of type FloatLike") 
        self.crater.rimwidth = np.float64(value)
        
    @property
    def peakheight(self) -> np.float64:
        """
        The height of the central peak in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.peakheight
    
    @peakheight.setter
    def peakheight(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("peakheight must be of type FloatLike") 
        self.crater.peakheight = np.float64(value)

    @property
    def floordiam(self) -> np.float64:
        """
        The diameter of the crater floor in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.floordiam
    
    @floordiam.setter
    def floordiam(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("floordiam must be of type FloatLike")
        self.crater.floordiam = np.float64(value)
        
    @property
    def floordepth(self) -> np.float64:
        """
        Return the depth of the crater floor in m
        
        Returns
        -------
        np.float64
        """
        return self.crater.floordepth
    
    @floordepth.setter
    def floordepth(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("floordepth must be of type FloatLike")
        self.crater.floordepth = np.float64(value)
        
    @property
    def ejrim(self) -> np.float64:
        """
        The thickness of ejecta at the rim in m.
        
        Returns
        -------
        np.float64
        """
        return self.crater.ejrim
    
    @ejrim.setter
    def ejrim(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("ejrim must be of type FloatLike")
        self.crater.ejrim = np.float64(value)
        
    @property
    def crater(self):
        """
        The crater to be created.
        
        Returns
        -------
        Crater
        """ 
        return self._crater
    
    @crater.setter
    def crater(self, value):
        from .impact import Crater
        if value is not None and not isinstance(value, Crater):
            raise TypeError("crater must be an instance of Crater")
        self._crater = value
        return 
        
    @property
    def target(self):
        """
        The target body for the impact.
        
        Returns
        -------
        Target
        """ 
        return self._target
    
    @target.setter
    def target(self, value):
        from .target import Target
        if value is None:
            self._target = Target(name="Moon")
            return
        if not isinstance(value, Target):
            raise TypeError("target must be an instance of Target")
        self._target = value
        return        
        
    @property
    def rng(self):
        """
        A random number generator instance.
        
        Returns
        -------
        Generator
        """ 
        return self._rng
    
    @rng.setter
    def rng(self, value):
        if not isinstance(value, Generator) and value is not None:
            raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
        self._rng = value or np.random.default_rng()   
       
    @property
    def ejecta_truncation(self) -> np.float64:
        """
        The radius at which the crater is truncated relative to the crater radius.
        
        Returns
        -------
        np.float64 or None
        """
        return self.crater.ejecta_truncation 
    
    @ejecta_truncation.setter
    def ejecta_truncation(self, value: FloatLike | None): 
        if value:
            if not isinstance(value, FloatLike):
                raise TypeError("truction_radius must be of type FloatLike")
            self.crater.ejecta_truncation = np.float64(value)
        else:
            self.crater.ejecta_truncation = None


import unittest
from unittest.mock import patch, MagicMock
from cratermaker import to_config, set_properties, check_properties, create_catalogue, validate_and_convert_location, normalize_coords  
from cratermaker.utils.custom_types import FloatLike

mock_properties = [
    "name",       "property1",     "property2"
]
mock_values = [
    ("foo",     1.00,     2.00),
    ("bar",     3.00,     4.00),
]  
class TestSetProperties(unittest.TestCase):
    def setUp(self):
        # Set up any initial configurations or objects you need before each test
        self.mock_object = MagicMock()
        return
        
    def test_kwargs(self):
        set_properties(self.mock_object, name="baz", property1=5.0, property2=6.0)
        self.assertEqual(self.mock_object.name, "baz")
        self.assertEqual(self.mock_object.property1, 5.0)
        self.assertEqual(self.mock_object.property2, 6.0)
       
        # Reset the properties and try to cause an exception when a name property is not passed 
        with self.assertRaises(ValueError): 
            del self.mock_object.name  # Explicitly delete the name attribute
            set_properties(self.mock_object, property1=5.0, property2=6.0) # name omitted
        return    
    
    def test_catalogue(self):
        # Test a regular catalogue with multiple entries (must be selected by  name)
        catalogue = create_catalogue(mock_properties, mock_values)
        for k,v in catalogue.items():
            set_properties(self.mock_object, catalogue=catalogue, name=k)
            self.assertEqual(self.mock_object.name, k)
            self.assertEqual(self.mock_object.property1, v['property1'])
            self.assertEqual(self.mock_object.property2, v['property2'])
            
        # Test a single entry catalogue (name can be omitted)
        name = next(iter(catalogue))
        set_properties(self.mock_object, catalogue={name:catalogue[name]})
    
        # Create an invalid catalogue structure
        invalid_catalogue = {"invalid_key": "invalid_value"}
        
        # Test if ValueError is raised for invalid catalogue structure or other problems
        with self.assertRaises(ValueError):  
            set_properties(self.mock_object, catalogue="bad_value") # Not even a dict
        with self.assertRaises(ValueError):  
            set_properties(self.mock_object, catalogue=invalid_catalogue, name="invalid_key") # Not a nested dict
        with self.assertRaises(ValueError):  
            set_properties(self.mock_object, catalogue=catalogue) # Name omitted
        
        return

    @patch('builtins.open', create=True)  # Adjust 'your_module' accordingly
    def test_json_file_setting(self, mock_open):
        # Mock reading from a JSON file
        mock_open.return_value.__enter__.return_value.read.return_value = '{"qux":{"property1":7.0,"property2":8.0},"quux":{"property1":9.0,"property2":10.0}}'
        set_properties(self.mock_object, filename='dummy.json', name='qux')
        self.assertEqual(self.mock_object.name, 'qux')
        self.assertEqual(self.mock_object.property1, 7.0)
        self.assertEqual(self.mock_object.property2, 8.0)
        set_properties(self.mock_object, filename='dummy.json', name='quux')
        self.assertEqual(self.mock_object.name, 'quux')
        self.assertEqual(self.mock_object.property1, 9.0)
        self.assertEqual(self.mock_object.property2, 10.0)
        return 
        
    def test_kwarg_override(self):
        property2_new = 42.0
        catalogue = create_catalogue(mock_properties, mock_values)
        name = next(iter(catalogue))
        set_properties(self.mock_object, catalogue=catalogue, name=name, property2=property2_new)
        self.assertEqual(self.mock_object.name, name)
        self.assertEqual(self.mock_object.property1, catalogue[name]['property1'])
        self.assertEqual(self.mock_object.property2, property2_new)
        
        # test that overrides are ignored when None is passed:
        set_properties(self.mock_object, catalogue=catalogue, name=name, property2=None) 
        self.assertEqual(self.mock_object.property2, catalogue[name]['property2'])
        return

if __name__ == '__main__':
    unittest.main()



import unittest
from cratermaker import Target, Material, Scale
import numpy as np

class TestScale(unittest.TestCase):
   
    def test_material_from_catalogue(self):
        # Test creating a material from the catalogue
        material = Material("Soft Rock")
        self.assertEqual(material.K1, 0.20)
        self.assertEqual(material.mu, 0.55)
        self.assertEqual(material.Ybar, 7.60e6)
        self.assertEqual(material.density, 2250.0)
        self.assertIsInstance(material.K1, np.float64)
        self.assertIsInstance(material.mu, np.float64)
        self.assertIsInstance(material.Ybar, np.float64)
        self.assertIsInstance(material.density, np.float64)
        return

    def test_material_override_catalogue(self):
        # Test overriding a property from the catalogue
        material = Material(name="Soft Rock", Ybar=0.0)
        self.assertEqual(material.Ybar, 0.0)
        return

    def test_custom_material(self):
        # Test creating a custom material
        material = Material(name="Flubber", K1=3.8, mu=0.1, Ybar=1e7, density=2000.0)
        self.assertEqual(material.K1, 3.8)
        self.assertEqual(material.mu, 0.1)
        self.assertEqual(material.Ybar, 1e7)
        self.assertEqual(material.density, 2000.0)
        return

    def test_incomplete_custom_material(self):
        # Test incomplete custom material creation
        with self.assertRaises(ValueError):
            Material("Flubber", density=2000.0)
        with self.assertRaises(ValueError):
            Material(K1=3.8, mu=0.1, Ybar=1e7, density=2000.0)
        return
    
    def test_scale_override_catalogue(self):
        # Test overriding a property from the catalogue
        target = Target(name="Mars")
        scale = Scale(target=target,material_name="Sand")
        self.assertEqual(scale.material.name, "Sand")    

if __name__ == '__main__':
    unittest.main()


import numpy as np
import os
import xarray as xr
from pathlib import Path
from dataclasses import dataclass, field, asdict
from typing import Tuple, Dict, Sequence, Optional, Any
import os
from ..utils.general_utils import set_properties, create_catalogue, check_properties
from ..utils.custom_types import FloatLike
import warnings

class Target:
    """
    Represents the target body in a crater simulation.

    This class encapsulates the properties of the target that is impacted, including
    its material composition, size, and other relevant physical characteristics.
    """
       
    config_ignore = ['catalogue','material']  # Instance variables to ignore when saving to file
    def __init__(self, 
                 name: str, 
                 radius: FloatLike | None = None, 
                 diameter: FloatLike | None = None, 
                 gravity: FloatLike | None = None, 
                 transition_scale_type: str | None = None,
                 material_name: str | None = None,                  
                 catalogue: Dict[str, Dict[str, FloatLike]] | None = None,
                 **kwargs: Any,
                 ):
        """
        Initialize the target object, setting properties from the provided arguments,
        and creating a catalogue of known solar system targets if not provided.

        Parameters
        ----------
        name : str or None
            Name of the target body.
        radius : FloatLike or None
            Radius of the target body in meters.
        diameter : FloatLike or None
            Diameter of the target body in meters.
        gravity : FloatLike or None
            Surface gravity of the target body in m/s^2.
        material_name : str or None
            Name of the material composition of the target body.            
        transition_scale_type : str or None
            Simple-to-complex transition scaling to use for the surface (either "silicate" or "ice").
        catalogue : dict or None
            Optional input of catalogue solar system targets to replace the built-in catalogue.
        **kwargs : Any
            Additional keyword argumments that could be set by the user.
        """    

        # Set the attributes for this class
        self._name = name
        self._gravity = None
        self._transition_scale_type = None
        self._material_name = None
        
        # ensure that only either diamter of radius is passed
        values_set = sum(x is not None for x in [diameter, radius])
        if values_set > 1:
            raise ValueError("Only one of diameter, radius may be set")        
        
        self._radius = None
        self._diameter = None
        self._catalogue = None
        self.catalogue = catalogue 
            
        # Set properties for the Target object based on the arguments passed to the function
        self.set_properties(name=name, 
                            radius=radius, 
                            diameter=diameter, 
                            gravity=gravity, 
                            material_name=material_name,
                            transition_scale_type=transition_scale_type, 
                            catalogue=self.catalogue,
                            **kwargs)

        check_properties(self) 
        return
    
    @property
    def catalogue(self):
        """
        A nested dictionary containing a catalogue of known solar system targets to use for the simulation. 
        
        Returns
        -------
        Dict[str, Dict[str, FloatLike]] or None
            A catalogue of known solar system targets to use for the simulation.
        """
        return self._catalogue

    @catalogue.setter
    def catalogue(self, value):
        
        if not isinstance(value, dict) and value is not None:
            raise TypeError("catalogue must be a dict or None")
        
        # Define some built-in catalogue values for known solar system targets of interest
        gEarth = np.float64(9.80665) # 1 g in SI units
        body_properties = [
            "name",    "radius",   "gravity",      "material_name", "transition_scale_type"
        ]
        body_values = [
            ("Mercury", 2440.0e3,  0.377 * gEarth, "Soft Rock", "silicate"),
            ("Venus",   6051.84e3, 0.905 * gEarth, "Hard Rock", "silicate"),
            ("Earth",   6371.01e3, 1.000 * gEarth, "Wet Soil" , "silicate"),
            ("Moon",    1737.53e3, 0.1657* gEarth, "Soft Rock", "silicate"),
            ("Mars",    3389.92e3, 0.379 * gEarth, "Soft Rock", "silicate"),
            ("Ceres",   469.7e3,   0.029 * gEarth, "Ice"      , "ice"),
            ("Vesta",   262.7e3,   0.025 * gEarth, "Soft Rock", "silicate"),
        ]
        
        if value is None: 
            self._catalogue = create_catalogue(body_properties, body_values)
        else:
            self._catalogue = value    
    

    def set_properties(self, **kwargs):
        """
        Set properties of the current object based on the provided keyword arguments.

        This function is a utility to update the properties of the current object. The actual implementation of the 
        property setting is handled by the `util.set_properties` method.

        Parameters
        ----------
        **kwargs : dict
            A dictionary of keyword arguments that represent the properties to be set on the current object.

        Returns
        -------
        None
            The function does not return a value.
        """         
        set_properties(self,**kwargs)
        return

    
    @property
    def escape_velocity(self):
        """
        Calculate the escape velocity for the target body.

        Returns
        -------
        np.float64
            Escape velocity in m/s.
        """        
        return np.sqrt(2 * self.radius * self.gravity)

    
    @property
    def name(self):
        """
        The name of the target body.
        
        Returns
        -------
        str 
        """
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("name must be a string or None")
        self._name = value

       
    @property
    def radius(self) -> Optional[float]:
        """
        Radius of the target body in m.
        
        Returns
        -------
        np.float64 
        """        
        return self._radius

    @radius.setter
    def radius(self, value: float):
        if not isinstance(value, float) or value <= 0:
            raise ValueError("Radius must be a positive number.")
        self._radius = np.float64(value)
        self._diameter = np.float64(value) * 2

    @property
    def diameter(self) -> Optional[float]:
        """
        Diameter of the target body in m.
        
        Returns
        -------
        np.float64 
        """                
        return self._diameter

    @diameter.setter
    def diameter(self, value: float):
        if not isinstance(value, float) or value <= 0:
            raise ValueError("Diameter must be a positive number.")
        self._diameter = np.float64(value)
        self._radius = np.float64(value) / 2

    @property
    def gravity(self):
        """
        Surface gravity of the target body in m/s^2.
        
        Returns
        -------
        np.float64
        """
        return self._gravity

    @gravity.setter
    def gravity(self, value):
        if value is not None and not isinstance(value, FloatLike):
            raise TypeError("gravity must be a numeric value or None")
        self._gravity = np.float64(value)

    @property
    def material_name(self):
        """
        The name of the material composition of the target body.
        
        Returns
        -------
        str 
        """
        return self._material_name

    @material_name.setter
    def material_name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("material_name must be a string or None")
        self._material_name = value
        
    @property
    def transition_scale_type(self):
        """
        The type of simple-to-complex transition scaling used for the surface, either 'silicate' or 'ice'.
        
        Returns
        -------
        str 
        """
        return self._transition_scale_type

    @transition_scale_type.setter
    def transition_scale_type(self, value):
        valid_types = ["silicate", "ice"]
        if value not in valid_types and value is not None:
            raise ValueError(f"Invalid transition_scale_type: {value}. Must be one of {valid_types} or None")
        self._transition_scale_type = value
